module cpu
(
    input               clock,
    input               reset_n,
    input               locked,
    /* verilator lint_off UNDRIVEN */
    output      [19:0]  address,
    input       [ 7:0]  i_data,
    output  reg [ 7:0]  o_data,
    output  reg         we
);

`include "decl.v"

always @(posedge clock)
// Либо сброс, либо инициализация
if (!reset_n || !locked) begin cs <= 16'hF000; ip <= 16'hFFF0; t <= init; end
else case (t)

    // Первый такт сбрасывает инструкцию перед запуском
    init: begin

        t       <= fetch;
        we      <= 1'b0;
        src     <= 1'b0;
        segment <= ds;
        segover <= 1'b0;
        rep     <= 2'b00;

    end

    // Считывание опкода
    fetch: begin

        ip      <= ip + 1'b1;
        opcode  <= i_data;
        size    <= i_data[0];
        dir     <= i_data[1];

        /* verilator lint_off CASEX */
        casex (i_data)

            // Обработка префиксов
            8'h26: begin segment <= es; segover <= 1'b1; end
            8'h2E: begin segment <= cs; segover <= 1'b1; end
            8'h36: begin segment <= ss; segover <= 1'b1; end
            8'h3E: begin segment <= ds; segover <= 1'b1; end
            8'hF2, 8'hF3: begin rep <= i_data[1:0]; end
            8'hF0, 8'h64, 8'h65, 8'h66, 8'h67, 8'h0F: begin /* ничего не делать */ end
            // Основное АЛУ
            8'b00_xxx_0xx: begin

                t   <= tmodrm;
                alu <= i_data[5:3];

            end
            // Запись опкода или частичное исполнение
            default: t <= exec;

        endcase

    end

    // Считывание байта modrm
    tmodrm: begin

        modrm   <= i_data;
        ip      <= ip + 1'b1;

        // Левый операнд (op1) регистр
        case (dir ? i_data[5:3] : i_data[2:0])

            3'b000: op1 <= size ? ax : ax[ 7:0];
            3'b001: op1 <= size ? cx : cx[ 7:0];
            3'b010: op1 <= size ? dx : dx[ 7:0];
            3'b011: op1 <= size ? bx : bx[ 7:0];
            3'b100: op1 <= size ? sp : ax[15:8];
            3'b101: op1 <= size ? bp : cx[15:8];
            3'b110: op1 <= size ? si : dx[15:8];
            3'b111: op1 <= size ? di : bx[15:8];

        endcase

        // Правый операнд (op2) регистр
        case (dir ? i_data[2:0] : i_data[5:3])

            3'b000: op2 <= size ? ax : ax[ 7:0];
            3'b001: op2 <= size ? cx : cx[ 7:0];
            3'b010: op2 <= size ? dx : dx[ 7:0];
            3'b011: op2 <= size ? bx : bx[ 7:0];
            3'b100: op2 <= size ? sp : ax[15:8];
            3'b101: op2 <= size ? bp : cx[15:8];
            3'b110: op2 <= size ? si : dx[15:8];
            3'b111: op2 <= size ? di : bx[15:8];

        endcase

        // Вычислить базовый эффективный адрес
        case (i_data[2:0])

            3'b000: ea <= bx + si;
            3'b001: ea <= bx + di;
            3'b010: ea <= bp + si;
            3'b011: ea <= bp + di;
            3'b100: ea <= si;
            3'b101: ea <= di;
            3'b110: ea <= ^i_data[7:6] ? bp : 1'b0;
            3'b111: ea <= bx;

        endcase

        // Выбор сегмента SS: по умолчанию, если не задан сегментный префикс
        // - Если младшие 3 бита равны 01x
        // - Или при mod=01,10 младшие биты равны 110
        if (!segover && (i_data[2:1] == 2'b01 || (^i_data[7:6] && i_data[2:0] == 3'b110)))
            segment <= ss;

        // Считывать +d8/+d16 или нет
        case (i_data[7:6])

            2'b00: begin

                // Читать +disp16
                if (i_data[2:0] == 3'b110) t <= tmodrm2;
                // Сразу читать операнды из памяти
                else begin t <= tmodrmf; src <= 1'b1; end

            end
            2'b01: t <= tmodrm1;
            2'b10: t <= tmodrm2;
            2'b11: t <= exec;

        endcase

    end

    // Чтение +disp8 и переход к чтению операнда из памяти
    tmodrm1: begin

        t   <= tmodrmf;
        ip  <= ip + 1'b1;
        ea  <= ea + {{8{i_data[7]}}, i_data[7:0]};
        src <= 1'b1;

    end

    // Младший байт disp16
    tmodrm2: begin

        t   <= tmodrmh;
        ip  <= ip + 1'b1;
        ea  <= ea + i_data;

    end

    // Старший байт disp16
    tmodrmh: begin

        t   <= tmodrmf;
        ip  <= ip + 1'b1;
        src  <= 1'b1;
        ea[15:8] <= ea[15:8] + i_data;

    end

    // Чтение первого байта из памяти
    tmodrmf: begin

        if (dir) op2 <= i_data; else op1 <= i_data;

        // Если это двухбайтный
        if (size)
             begin t <= tmodrmf2; ea <= ea + 1'b1; end
        else begin t <= exec; end

    end

    // Чтение второго байта
    tmodrmf2: begin

        if (dir) op2[15:8] <= i_data; else op1[15:8] <= i_data;

        // Вернуть на место ea
        ea <= ea - 1'b1;
        t  <= exec;

    end

endcase

endmodule
